#ifndef CLONE_STRUCT_H
#define CLONE_STRUCT_H

typedef enum CLONE_BOOL {
  CLONE_TRUE=1,
  CLONE_FALSE=0
} CLONE_BOOL;

typedef enum CLONE_STATE {
  CLONE_IDLE,
  CLONE_SPEAKING
} CLONE_STATE;

typedef enum CLONE_MODEL {
  CLONE_MODEL_HELENE,
  CLONE_MODEL_PIERRE,
  CLONE_MODEL_SASHA,
  CLONE_MODEL_YASMINE,
  CLONE_MODEL_YASMINE_LPC
} CLONE_MODEL;

typedef struct CLONE_POS {
  double x;
  double y;
  double z;
} CLONE_POS;


/* MUST BE LIKE A T3DEULER */
typedef struct CLONE_POS_CORR {
  double rz;  
  double ry;
  double rx;
  double x;
  double y;
  double z;
} CLONE_POS_CORR;


#define CLONE_STRING_LENGTH 256

typedef struct CLONE_INIT {
  CLONE_MODEL model;
  CLONE_BOOL pomOn;
  char displayEnv[CLONE_STRING_LENGTH];
} CLONE_INIT;


typedef struct CLONE_WINDOW_PARAMS {
  double scale;
  double x;
  double y;
} CLONE_WINDOW_PARAMS;

typedef struct CLONE_WINDOW_SIZES {
  double width;
  double height;
} CLONE_WINDOW_SIZES;

typedef struct CLONE_PLAY_FILE_PARAMS {
  char txtFile[CLONE_STRING_LENGTH];
  char synFiles[CLONE_STRING_LENGTH];
} CLONE_PLAY_FILE_PARAMS;

#define CLONE_MAX_SPEECH_LENGTH 1024
typedef struct CLONE_SPEECH {
  int id;
  int unused;
  char french[CLONE_MAX_SPEECH_LENGTH];
  char english[CLONE_MAX_SPEECH_LENGTH];
} CLONE_SPEECH;


typedef struct CLONE_SENTENCE {
  char french[CLONE_MAX_SPEECH_LENGTH];
  char english[CLONE_MAX_SPEECH_LENGTH];
} CLONE_SENTENCE;

typedef enum CLONE_FRAME {
  CLONE_CLONE_FRAME,
  CLONE_ROBOT_FRAME,
  CLONE_WORLD_FRAME
} CLONE_FRAME;

typedef enum CLONE_MOTION {
  CLONE_MOVES_HEAD,
  CLONE_MOVES_EYES,
  CLONE_MOVES_BOTH
} CLONE_MOTION;

typedef struct CLONE_LOOK_AT {
  CLONE_MOTION motion;
  CLONE_FRAME frame;
  CLONE_POS pos;
} CLONE_LOOK_AT;

typedef struct CLONE_LOOK_AT_POINT {
  CLONE_FRAME frame;
  int unused;
  CLONE_POS pos;
} CLONE_LOOK_AT_POINT;

typedef struct CLONE_TRACK {
  char posterName[CLONE_STRING_LENGTH];
  CLONE_MOTION motion;
  int unused;
} CLONE_TRACK;

typedef struct CLONE_SPEECH_OBJ {
  char name[CLONE_STRING_LENGTH];
  CLONE_POS pos;
} CLONE_SPEECH_OBJ;


typedef enum CLONE_STEREOTYPE_TYPE {
  CLONE_STEREOTYPE_YES,
  CLONE_STEREOTYPE_NO,
  CLONE_STEREOTYPE_WHEREAREYOU
} CLONE_STEREOTYPE_TYPE;

typedef struct CLONE_STEREOTYPE {
  CLONE_STEREOTYPE_TYPE type;
  int nbIter;
  double length;
} CLONE_STEREOTYPE;

#endif /* CLONE_STRUCT_H */
