/**
 ** cloneExecTaskCodels.c
 **
 ** Codels appeles par la tache cloneExecTask
 **
 ** Auteur: 
 ** Date: 
 **
 **/

#define VERBOSE 1

#include <portLib.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/timeb.h>
#include <string.h>
#include <unistd.h>

#include "h2evnLib.h"

#include "server/cloneHeader.h"

#include "cloneConst.h"

#include "lib/Minim_lib.h"

#include "pomStruct.h"
#include "libeuler/pomEuler.h"
#include <t3d/t3d.h>

static POSTER_ID pomPosId = NULL;

static  POM_EULER_V t3d_m2o;
static  POM_EULER_V t3d_c2s;
static  POM_EULER_V t3d_c2m;
static  POM_EULER_V t3d_m2c;
static  POM_EULER_V t3d_c2o;
static  POM_EULER_V t3d_o2c;
static int firsttime=1;
static  float robot_to_clone[16];
static  float world_to_clone[16];


/* Manage activity list */
static int nextActivity[MAX_ACTIVITIES];
static int headActivity=-1;
static int lastActivity=-1;
static void addThisActivityInFifo();
static BOOL removeThisActivityFromFifo();
static BOOL isThisActivityHead();
static BOOL isThisActivityInFifo();
static BOOL exportSpeech(char *filename);
static void buildFileNames(CLONE_PLAY_FILE_PARAMS *playFileParams,
			   char *wav, char *art, char *txt);

static void cloneFrameCorrection(CLONE_POS *pos, CLONE_FRAME frame) ;
static BOOL cloneUpdateClonePositions (int *bilan);

static void  cloneSetMinimPos(POM_SENSOR_POS *screenPos); 

static BOOL cloneMove(CLONE_POS *pos, CLONE_MOTION motion, CLONE_FRAME frame, 
		      int *bilan);

static void transfPt(CLONE_POS *oldPos, POM_EULER_V *t3dmatrix, CLONE_POS *newPos);
static BOOL changePtFrame(CLONE_POS *oldPos, CLONE_FRAME oldFrame,
			  CLONE_POS *newPos, CLONE_FRAME newFrame);

/*------------------------------------------------------------------------ */

static struct timeb prev_MINIM_END_TIME;

static CLONE_SPEECH_POSTER_STR speechPoster;

static  char art[CLONE_MAX_SPEECH_LENGTH];
static  char wav[CLONE_MAX_SPEECH_LENGTH];
static  char txt[CLONE_MAX_SPEECH_LENGTH];

#define PIX_TO_HORI(p) ((p)*CLONE_SCREEN_WIDTH/CLONE_SCREEN_RESOLUTION_HORI)
#define PIX_TO_VERT(p) ((p)*CLONE_SCREEN_HIGH/CLONE_SCREEN_RESOLUTION_VERT)

/*------------------------------------------------------------------------
 *
 * cloneExecTaskInit  --  Codel d'initialisation (SDI/f, ...)
 *
 * Description: 
 * 
 * Retourne:    OK ou ERROR
 */

STATUS cloneExecTaskInit(int *bilan)
{
  int i;

  SDI_F->initParams.model = CLONE_MODEL_HELENE;
  SDI_F->initParams.pomOn = CLONE_FALSE;
  SDI_F->cloneThreadOn = CLONE_FALSE;
  SDI_F->cloneState = CLONE_IDLE;
  SDI_F->hideWindow = CLONE_FALSE;

  SDI_F->lookCorrection.rz = CLONE_DEFAULT_CORR_RZ;
  SDI_F->lookCorrection.ry = CLONE_DEFAULT_CORR_RY;
  SDI_F->lookCorrection.rx = CLONE_DEFAULT_CORR_RX;
  SDI_F->lookCorrection.x = CLONE_DEFAULT_CORR_X;
  SDI_F->lookCorrection.y = CLONE_DEFAULT_CORR_Y;
  SDI_F->lookCorrection.z = CLONE_DEFAULT_CORR_Z;

  SDI_F->windowParams.scale = CLONE_DEFAULT_POS_SCALE;
  SDI_F->windowParams.x = CLONE_DEFAULT_POS_X;
  SDI_F->windowParams.y = CLONE_DEFAULT_POS_Y;

  SDI_F->windowSizeInit.width = 0;
  SDI_F->windowSizeInit.height = 0;

  strcpy (SDI_F->playFileParams.txtFile, CLONE_DEFAULT_TXT_FILE);
  strcpy (SDI_F->playFileParams.synFiles, CLONE_DEFAULT_SYN_FILES);

/*   strcpy (SDI_F->playFileParams.artFileName, CLONE_DEFAULT_ART_FILE_NAME); */
/*   strcpy (SDI_F->playFileParams.wavFileName, CLONE_DEFAULT_WAV_FILE_NAME); */

  for(i=0; i<MAX_ACTIVITIES; i++) 
    nextActivity[i]=-1;

  return OK;
}



/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_clone_BAD_PARAM
 *              S_clone_INIT_CLONE_FAILED
 */

/* cloneInitExec  -  codel EXEC de Init
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneInitExec (CLONE_INIT *init, int *bilan)
{
  static int count = 0;
  char displayEnv[64];

  if (strlen(init->displayEnv)>2) {
    sprintf(displayEnv, "DISPLAY=%s", init->displayEnv);
    printf ("clone: putenv(%s)\n", displayEnv);
    putenv(displayEnv);
  }

  if (!SDI_F->cloneThreadOn) {
    /* the module must run in the data/ directory */
    chdir (CLONE_DATA_PATH);
    
    /* set clone definition */
    switch (init->model) {
    case CLONE_MODEL_PIERRE:
      Minim_def_cloneS("-mdl mdl/pb_default.txt -tex tex/pb_tex_3.txt -msh msh/pb_mobile.txt -back img/pb_tex_upu_face_visio.pnm -mov mov/pb_tex_upu_face.txt -cam cam/pb_labio.txt -mach dents/pb_mach_ili.txt -pal dents/pb_palais.txt -art art/null.txt");
      break;
      
    case CLONE_MODEL_HELENE:
      /*     Minim_def_cloneS("-back #001E32 -mach dents/hl_mach_tel.txt -pal dents/hl_palais_tel.txt -tex tex/hl_cyl_2_full.txt -msh msh/hl_full.txt -mdl mdl/hl_tts.txt -eye eye/hl.mod eye/null.art"); */
      Minim_def_cloneS("-back #001E32 -mach dents/hl_mach_tel.txt -pal dents/hl_palais_tel.txt -tex tex/hl_cyl_2_full.txt -msh msh/hl_full.txt -mdl mdl/hl_tts.txt -eye eye/hl_tex.mod eye/null.art -art art/null.txt  -cam cam/hl_default.txt");
      break;

    case CLONE_MODEL_SASHA:
      Minim_def_cloneS("-back #001E32 -obj Mdl2 mod/sf_oeil.mod -eye eye/sf_tex.mod eye/null.art -cam cam/sf_default.txt -msh msh/sf_oeil_plus.txt -tex tex/sf_rot_a_neu.txt -art art/null.txt");
      break;
      
    case CLONE_MODEL_YASMINE:
      Minim_def_cloneS("-back #001E32 -cam cam/yb_visage.txt -obj Hand mod/ybadsi_visRT_HDfixe.mod msh/ybb_full.msh tex/yb_cyl_rot_i.txt -mach dents/ybb_hl_mach_tel.txt -pal dents/ybb_hl_palais_tel.txt -mdl mdl/pb_default.txt -msh msh/null.txt -art art/null.txt");
      break;

    default:
      *bilan = S_clone_INIT_CLONE_FAILED;
      return ETHER;
    }
    
    /* enable clone launch */
    SDI_F->cloneThreadOn = CLONE_TRUE;
    
    /* set screen resolution */
/*     Minim_set_screen_resolution(CLONE_SCREEN_RESOLUTION_HORI,  */
/* 				CLONE_SCREEN_WIDTH, */
/* 				CLONE_SCREEN_RESOLUTION_VERT, */
/* 				CLONE_SCREEN_HIGH); */
  }

  /* wake up minim thread */
  h2evnSignal(EXEC_TASK_ID(CLONE_CLONETHREAD_NUM));

  if (MINIM_NOT_READY && count < 10) {
    count++;
    return EXEC;
  }
  if (MINIM_NOT_READY) {
    printf ("cloneInitExec: MINIM STILL NOT READY ?!\n");
    return ETHER;
  }

  /* init screen pos */
  if (!cloneUpdateClonePositions (bilan)) 
    return ETHER;

  /* pos the window */
  cloneSetWindowGeomExec(&SDI_F->windowParams, bilan);

  return ETHER;
}

/* ----------------------------------------------------------------------
 *
 *  cloneSetMinimPos
 *
 *
 */

static void  cloneSetMinimPos(POM_SENSOR_POS *screenPos)
{
  POM_EULER_V cloneToScreen;
  int i;
  double cloneInScreenHori, cloneInScreenVert;
  T3D mat_t3d_m2c;
  T3D mat_t3d_o2c;

  /* Position of the clone in the screen.
     The frame of the screen is at the middle of the screen with Z comming out 
     from the screen, X on the rigth, and Y upwards. */
  cloneInScreenHori = - CLONE_SCREEN_WIDTH / 2.0 
			 + PIX_TO_HORI(SDI_F->windowParams.x 
					  + MINIM_WIDTH/2.0*SDI_F->windowParams.scale);
  cloneInScreenVert = CLONE_SCREEN_HIGH / 2.0 
			 - PIX_TO_VERT(SDI_F->windowParams.y
					  + MINIM_HEIGHT/2.0*SDI_F->windowParams.scale);
		
#if VERBOSE >= 0 
  if (firsttime)
    printf ("clone in screen (m) : H %5.4f    V %5.4f\n",
	    cloneInScreenHori, cloneInScreenVert);
#endif

  /* make clone to screen (c2s) matrix */
  cloneToScreen.euler.yaw = 0.0;
  cloneToScreen.euler.pitch = 0.0;
  cloneToScreen.euler.roll = 0.0;
  cloneToScreen.euler.x = cloneInScreenHori;
  cloneToScreen.euler.y = cloneInScreenVert;
  cloneToScreen.euler.z = 0.0;
  t3d_c2s = cloneToScreen;

#if VERBOSE >= 0
  if (firsttime)
    printf ("clone  in screen (m) : x %5.4f  y %5.4f  z %5.4f\n",
	    t3d_c2s.euler.x,
	    t3d_c2s.euler.y,
	    t3d_c2s.euler.z);
#endif

  /* make screen to robot (s2m) matrix */
#if VERBOSE > 0    
  if (firsttime)
    printf ("screen in robot (m)  : x %5.4f  y %5.4f  z %5.4f\n",
	    screenPos->sensorToMain.euler.x,
	    screenPos->sensorToMain.euler.y,
	    screenPos->sensorToMain.euler.z);
#endif

  /* computes clone to robot (WARNING: s2m is now c2m) */
  pomEulerVCompIn(&t3d_c2m, &cloneToScreen, &screenPos->sensorToMain);

#if VERBOSE >= 0  
  if (firsttime)
    printf ("clone  in robot (m)  : x %5.4f  y %5.4f  z %5.4f\n",
	    t3d_c2m.euler.x,
	    t3d_c2m.euler.y,
	    t3d_c2m.euler.z);
#endif

  /* make robot to world (m2o) */
  t3d_m2o = screenPos->mainToOrigin;

#if VERBOSE > 0    
  if (firsttime)
    printf ("robot  in world (m)  : x %5.4f  y %5.4f  z %5.4f\n",
	    t3d_m2o.euler.x,
	    t3d_m2o.euler.y,
	    t3d_m2o.euler.z);
#endif

  /* clone to world (WARNING, m2o is now c2o) */
  pomEulerVCompIn(&t3d_c2o, &t3d_c2m, &screenPos->mainToOrigin);

#if VERBOSE > 0    
  if (firsttime)
    printf ("clone  in world (m)  : x %5.4f  y %5.4f  z %5.4f\n",
	    t3d_c2o.euler.x,
	    t3d_c2o.euler.y,
	    t3d_c2o.euler.z);
#endif

  /* world (origin) to clone */
  t3d_o2c = t3d_c2o;
  pomEulerVInvert(&t3d_o2c);

  /* robot (main) to clone */
  t3d_m2c = t3d_c2m;
  pomEulerVInvert(&t3d_m2c);

  /* get back to matrix */
  pomEulerVToT3D(&t3d_m2c, &mat_t3d_m2c);
  pomEulerVToT3D(&t3d_o2c, &mat_t3d_o2c);

  t3dConvertTo(T3D_MATRIX, &mat_t3d_m2c);
  t3dConvertTo(T3D_MATRIX, &mat_t3d_o2c);

  for (i=0; i<16; i++) {
    robot_to_clone[i] = mat_t3d_m2c.matrix.matrix[i];
    world_to_clone[i] = mat_t3d_o2c.matrix.matrix[i];
  }

/*   Minim_robot_to_clone(robot_to_clone); */
/*   Minim_world_to_clone(world_to_clone); */

  firsttime = 0;
}

/* ----------------------------------------------------------------------
 * clone get robot  position
 */
static BOOL
cloneUpdateClonePositions (int *bilan)
{
  POM_SENSOR_POS screenPos;
  
  /* not pom : hard code screen pos */
  /*    BECARFUL : at least be coherente with pom/model/b21r.geom */
  if (!SDI_F->initParams.pomOn) {
  
    screenPos.sensorToMain.euler.yaw = CLONE_SCREEN_IN_ROBOT_ROT_Z;
    screenPos.sensorToMain.euler.pitch = CLONE_SCREEN_IN_ROBOT_ROT_Y;
    screenPos.sensorToMain.euler.roll = CLONE_SCREEN_IN_ROBOT_ROT_X;
    screenPos.sensorToMain.euler.x = CLONE_SCREEN_CENTER_IN_ROBOT_X;
    screenPos.sensorToMain.euler.y = CLONE_SCREEN_CENTER_IN_ROBOT_Y;
    screenPos.sensorToMain.euler.z = CLONE_SCREEN_CENTER_IN_ROBOT_Z;
    
    screenPos.mainToOrigin.euler.yaw = 0;
    screenPos.mainToOrigin.euler.pitch = 0;
    screenPos.mainToOrigin.euler.roll = 0;
    screenPos.mainToOrigin.euler.x = 0;
    screenPos.mainToOrigin.euler.y = 0;
    screenPos.mainToOrigin.euler.z = 0;
  }

  /* robot position with pom */
  else {
    /* Init */
    if (pomPosId == NULL) {
      /* Set the pomPosId */
      if (posterFind("pomSreenFramePos", &pomPosId) != OK) {  
	pomPosId = NULL;
	printf ("cloneReadRobotPosition: cannot find screenPos poster pomScreenFramePos\n");
	*bilan = S_clone_CANNOT_CONNECT_TO_POM;
	return FALSE;
      } else {
	printf("clone : connected to screenPos poster\n");
      }
    } 
    
    /* Get the current position estimate */
    if (pomPosId != NULL)
      if(clonePOM_SENSOR_POSPosterRead(pomPosId, &screenPos) != OK) {
	printf ("cloneReadRobotPosition:  cannot read  screenPos poster pomScreenFramePos\n");
	pomPosId = NULL;
	*bilan = S_clone_CANNOT_CONNECT_TO_POM;
	return FALSE;
      }
  }

  if (firsttime)
    printf ("mainToOrigin pos : %g %g %g   ang :%g %g %g\n", 
	    screenPos.mainToOrigin.euler.x,
	    screenPos.mainToOrigin.euler.y,
	    screenPos.mainToOrigin.euler.z,
	    screenPos.mainToOrigin.euler.yaw,
	    screenPos.mainToOrigin.euler.pitch,
	    screenPos.mainToOrigin.euler.roll);

#if 0
  /* !!!!!! pour debug !!!!!!!*/
  printf("cloneUpdateClonePositions : current\n");
  for(i=0;i<6;i++)
    printf("sensorToMain Euler[%d] = %f\n",i,screenPos.sensorToMain.euler[i]);
  for(i=0;i<6;i++)
    printf("mainToBase Euler[%d] = %f\n",i,screenPos.mainToBase.euler[i]);
  for(i=0;i<6;i++)
    printf("mainToOrigin Euler[%d] = %f\n",i,screenPos.mainToOrigin.euler[i]);
#endif


  /* clone set screen pos */
  cloneSetMinimPos(&screenPos);

  /* ok */
  return TRUE;
}

/*------------------------------------------------------------------------
 *
 * cloneExecTaskEnd  --  Codel de terminaison
 *
 * Description: 
 * 
 * Retourne:    OK ou ERROR
 */

STATUS cloneExecTaskEnd(void)
{
  Minim_finish_clone();
  return OK;
}


/*------------------------------------------------------------------------
 * SetHideWindow
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* cloneSetHideWindowExec  -  codel EXEC de SetHideWindow
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneSetHideWindowStart (CLONE_BOOL *hideWindow, int *bilan)
{  
  cloneSetWindowGeomExec(&SDI_F->windowParams, bilan);
  return EXEC;
}
ACTIVITY_EVENT cloneSetHideWindowExec (CLONE_BOOL *hideWindow, int *bilan)
{  
  if (*hideWindow)
    Minim_hide_clone();
  else {
    Minim_show_clone();
  }
  return END;
}
ACTIVITY_EVENT cloneSetHideWindowEnd (CLONE_BOOL *hideWindow, int *bilan)
{  
  cloneSetWindowGeomExec(&SDI_F->windowParams, bilan);
  return ETHER;
}

/*------------------------------------------------------------------------
 * SetWindowGeom
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* cloneSetWindowGeomExec  -  codel EXEC de SetWindowGeom
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneSetWindowGeomExec (CLONE_WINDOW_PARAMS *windowParams, int *bilan)
{
  if (MINIM_NOT_READY) {
    *bilan = S_clone_CLONE_NOT_INIT;
    return ETHER;
  }

  if(!SDI_F->windowSizeInit.height) {
    SDI_F->windowSizeInit.height = MINIM_HEIGHT;
    SDI_F->windowSizeInit.width = MINIM_WIDTH;
  }
  Minim_set_window_size((int)((double)SDI_F->windowSizeInit.width*windowParams->scale), 
			(int)((double)SDI_F->windowSizeInit.height*windowParams->scale));

  Minim_set_window_position(windowParams->x, windowParams->y);

  cloneUpdateClonePositions (bilan);

  return ETHER;
}


/*------------------------------------------------------------------------
 * PlayFile
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_clone_FAILED
 */

/* clonePlayFileStart  -  codel START de PlayFile
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT clonePlayFileStart (CLONE_PLAY_FILE_PARAMS *playFileParams, int *bilan)
{
  if (!isThisActivityInFifo())
    addThisActivityInFifo();

  /* not yet activity turn */
  if (!isThisActivityHead()) 
    return START;
  
  /* export text in poster */
  buildFileNames(playFileParams, wav, art, txt);
  exportSpeech(txt);

  printf ("clonePlayFileStart : lets go acti %d\n", 
	  CURRENT_ACTIVITY_NUM(CLONE_EXECTASK_NUM));

  /* let's go */
  /* record previous end time */
  prev_MINIM_END_TIME = MINIM_END_TIME;

  /* start speaking */
  Minim_run_sequence(art, wav, 1 /* remove files at end */);

  return EXEC;
}

/* clonePlayFileExec  -  codel EXEC de PlayFile
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT clonePlayFileExec (CLONE_PLAY_FILE_PARAMS *playFileParams, 
				  int *bilan)
{
  struct timeb timeOn;
  unsigned long int on;

  /* wait real start */
  if (prev_MINIM_END_TIME.time == MINIM_END_TIME.time &&
      prev_MINIM_END_TIME.millitm == MINIM_END_TIME.millitm)
    return EXEC;

  /* estimated duration */
  ftime(&timeOn);
  SDI_F->timeEnd = MINIM_END_TIME.time * 1000 + MINIM_END_TIME.millitm;
  on = timeOn.time * 1000 + timeOn.millitm;
  printf ("clonePlayFileExec acti %d for %ldms\n",
	  CURRENT_ACTIVITY_NUM(CLONE_EXECTASK_NUM),
	  SDI_F->timeEnd - on);

  return END;
}

/* clonePlayFileEnd  -  codel END de PlayFile
   Retourne:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT clonePlayFileEnd (CLONE_PLAY_FILE_PARAMS *playFileParams, 
				 int *bilan)
{
  /* wait end */
  if (MINIM_IS_SPEAKING) {

    return END;
  }

  /* its over */
  removeThisActivityFromFifo();

  return ETHER;
}

/* clonePlayFileInter  -  codel INTER de PlayFile
   Retourne:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT clonePlayFileInter (CLONE_PLAY_FILE_PARAMS *playFileParams, 
				 int *bilan)
{
  /* its over */
  removeThisActivityFromFifo();

  return ETHER;
}

/*------------------------------------------------------------------------
 * PlayRef
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_clone_FAILED
 */

/* clonePlayRefStart  -  codel START de PlayRef
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT clonePlayRefStart (int *playRefParams, int *bilan)
{
  if (!isThisActivityInFifo())
    addThisActivityInFifo();

  /* not yet activity turn */
  if (!isThisActivityHead()) 
    return START;
  
  /* let's go */
  /* record previous end time */
  prev_MINIM_END_TIME = MINIM_END_TIME;

  /*  */
  exportSpeech(txt);
  Minim_play_msg(*playRefParams);

  return EXEC;
}

/* clonePlayRefExec  -  codel EXEC de PlayRef
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT clonePlayRefExec (int *playRefParams, int *bilan)
{
  struct timeb timeOn;
  unsigned long int on;

  /* wait real start */
  if (prev_MINIM_END_TIME.time == MINIM_END_TIME.time &&
      prev_MINIM_END_TIME.millitm == MINIM_END_TIME.millitm)
    return EXEC;

  /* estimated duration */
  ftime(&timeOn);
  SDI_F->timeEnd = MINIM_END_TIME.time * 1000 + MINIM_END_TIME.millitm;
  on = timeOn.time * 1000 + timeOn.millitm;
  printf ("clonePlayFileExec acti %d for %ldms\n",
	  CURRENT_ACTIVITY_NUM(CLONE_EXECTASK_NUM),
	  SDI_F->timeEnd - on);

  return END;
}

/* clonePlayRefEnd  -  codel END de PlayRef
   Retourne:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT clonePlayRefEnd (int *playRefParams, int *bilan)
{
  /* wait end */
  if (MINIM_IS_SPEAKING) {

    return END;
  }

  /* its over */
  removeThisActivityFromFifo();

  return ETHER;
}

/* clonePlayRefInter  -  codel INTER de PlayRef
   Retourne:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT clonePlayRefInter (int *playRefParams, int *bilan)
{
  /* its over */
  removeThisActivityFromFifo();

  return ETHER;
}

/*------------------------------------------------------------------------
 * RePlay
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_clone_FAILED
 */


/* cloneRePlayExec  -  codel EXEC de RePlay
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneRePlayExec (int *bilan)
{
  exportSpeech(txt);
  Minim_play_msg(0);
  return ETHER;
}

/*------------------------------------------------------------------------
 * CleanFifo
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* cloneCleanFifoExec  -  codel EXEC de CleanFifo
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneCleanFifoExec (int *bilan)
{
  Minim_stop_msg();
  return ETHER;
}


/*------------------------------------------------------------------------
 * StopPlay
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* cloneStopPlayExec  -  codel EXEC de StopPlay
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneStopPlayExec (int *bilan)
{
  Minim_stop_msg();
  return ETHER;
}

/*------------------------------------------------------------------------
 * CleanFile
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_clone_UNKNOWN_FILE
 *              S_clone_CANNOT_CLEAN
 */

/* cloneCleanFileExec  -  codel EXEC de CleanFile
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneCleanFileExec (CLONE_PLAY_FILE_PARAMS *playFileParams, 
				   int *bilan)
{
  return ETHER;
}


/* ========================= LOCAL FUNCTIONS ========================= */


/*------------------------------------------------------------------------
 * addThisActivityInFifo
 *
 * Description: 
 *
 */

static void addThisActivityInFifo()
{
  int thisActivity = CURRENT_ACTIVITY_NUM(CLONE_EXECTASK_NUM);

  /* nothing on, let's go */
  if (headActivity == -1) {
    headActivity = thisActivity;
    nextActivity[thisActivity] = -1;
    lastActivity = thisActivity;
    printf ("new acti %d at head\n", thisActivity);
    return;
  }
  
  /* add at end of list */
  nextActivity[lastActivity] = thisActivity;
  lastActivity = thisActivity;
  nextActivity[thisActivity] =  -1;
  printf ("new acti %d in list\n", thisActivity);
}

static BOOL isThisActivityHead()
{
  int thisActivity = CURRENT_ACTIVITY_NUM(CLONE_EXECTASK_NUM);
  BOOL isHead;

  isHead = (headActivity == thisActivity);
  printf ("acti %d %s\n", thisActivity, isHead ? "ON" : "wait");

  return (isHead);
}

static BOOL isThisActivityInFifo()
{
  int thisActivity = CURRENT_ACTIVITY_NUM(CLONE_EXECTASK_NUM);
  BOOL inFifo;

  /* in list if next exists or if last of the list */
  inFifo = nextActivity[thisActivity] != -1 || lastActivity == thisActivity;

  printf ("acti %d %s\n", thisActivity, inFifo ? "recorded" : "unknown");

  return (inFifo);
}

static BOOL removeThisActivityFromFifo()
{
  int thisActivity = CURRENT_ACTIVITY_NUM(CLONE_EXECTASK_NUM);
  int acti;

  /* it is the current activity */
  if (thisActivity == headActivity) {

    /* it was the only activity */
    if (nextActivity[thisActivity] == -1) {
      headActivity = -1;
      lastActivity = -1;
    }
    /* there is a following one */
    else {
      headActivity = nextActivity[thisActivity];
      nextActivity[thisActivity] = -1;
    }
    return TRUE;
  }

  /* look for this activity in list */
  acti = headActivity;
  while (acti != -1 && nextActivity[acti] != thisActivity) {
    acti = nextActivity[acti];
  }

  /* not found */
  if (acti == -1) {
    printf ("activity %d not found\n", thisActivity);
    return FALSE;
  }

  /* found */
  nextActivity[acti] = nextActivity[thisActivity];
  nextActivity[thisActivity] = -1;
  if (thisActivity == lastActivity)
    lastActivity = -1;

  return TRUE;
}


/*------------------------------------------------------------------------
 * exportSpeech
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_clone_FAILED
 */

static BOOL exportSpeech(char *filename)
{
  FILE *pfile;
  int len;

  if (!(pfile = fopen(filename, "r"))) {
    printf ("clone: cannot open %s\n", filename);
    return FALSE;
  }

  fgets(speechPoster.Speech.french, CLONE_MAX_SPEECH_LENGTH, pfile);
  fgets(speechPoster.Speech.english, CLONE_MAX_SPEECH_LENGTH, pfile);

  fclose(pfile);

  speechPoster.Speech.id++;

  len = strlen(speechPoster.Speech.french);
  if ( speechPoster.Speech.french[len-1] == '\n' ) {
    speechPoster.Speech.french[len-1] = '\0';
  }

  len = strlen(speechPoster.Speech.english);
  if ( speechPoster.Speech.english[len-1] == '\n' ) {
    speechPoster.Speech.english[len-1] = '\0';
  }

  if (cloneSpeechPosterWrite(CLONE_SPEECH_POSTER_ID, &speechPoster) != OK) {
    printf ("clone exportSpeech: cloneSpeechPosterWrite fialed\n");
    return FALSE;
  }

  return TRUE;
}


/*------------------------------------------------------------------------
 * buildFileNames
 *
 * Description: 
 *
 */

static void buildFileNames(CLONE_PLAY_FILE_PARAMS *playFileParams,
			   char *wav, char *art, char *txt)
{
  if (playFileParams->synFiles[strlen(playFileParams->synFiles)-1] == '/')
    playFileParams->synFiles[strlen(playFileParams->synFiles)-1] = '\0';

  sprintf(wav, "%s.syn", playFileParams->synFiles);
  sprintf(art, "%s.syn.art", playFileParams->synFiles);
  sprintf(txt, "%s", playFileParams->txtFile);
}

/*------------------------------------------------------------------------
 * ExportSentence
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* cloneExportSentenceExec  -  codel EXEC de ExportSentence
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneExportSentenceExec (CLONE_SENTENCE *sentence, int *bilan)
{
  int length;

  length = MAX(strlen(sentence->french),
	       strlen(sentence->english));
  if (length > CLONE_MAX_SPEECH_LENGTH) {
    *bilan = S_clone_BAD_PARAMS;
    return ETHER;
  }

  strncpy(speechPoster.Speech.french, sentence->french,
	  CLONE_MAX_SPEECH_LENGTH);
  strncpy(speechPoster.Speech.english, sentence->english,
	  CLONE_MAX_SPEECH_LENGTH);
  speechPoster.Speech.id++;

  if (cloneSpeechPosterWrite(CLONE_SPEECH_POSTER_ID, &speechPoster) != OK) {
    printf ("clone exportSpeech: cloneSpeechPosterWrite failed\n");
    return ETHER;
  }
  return ETHER;
}


/*------------------------------------------------------------------------
 * LookAt
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* cloneLookAtExec  -  codel EXEC de LookAt
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneLookAtExec (CLONE_LOOK_AT *lookat, int *bilan)
{
  CLONE_LOOK_AT lookatTmp=*lookat;

  /* let's move */
  if (!cloneMove(&lookatTmp.pos, lookatTmp.motion, lookatTmp.frame, bilan)) 
    return ETHER;

  return EXEC;
}


/*------------------------------------------------------------------------
 * HeadLookAt
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* cloneHeadLookAtExec  -  codel EXEC de HeadLookAt
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneHeadLookAtExec (CLONE_LOOK_AT_POINT *lookat, int *bilan)
{
  CLONE_LOOK_AT_POINT lookatTmp=*lookat;

  /* let's move */
  if (!cloneMove(&lookatTmp.pos, CLONE_MOVES_HEAD, lookatTmp.frame, bilan)) 
    return ETHER;

  return EXEC;
}


/*------------------------------------------------------------------------
 * EyesLookAt
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* cloneEyesLookAtExec  -  codel EXEC de EyesLookAt
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneEyesLookAtExec (CLONE_LOOK_AT_POINT *lookat, int *bilan)
{
  CLONE_LOOK_AT_POINT lookatTmp=*lookat;

  /* let's move */
  if (!cloneMove(&lookatTmp.pos, CLONE_MOVES_EYES, lookatTmp.frame, bilan)) 
    return ETHER;

  return EXEC;
}


/*------------------------------------------------------------------------
 * HeadAndEyesLookAt
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* cloneHeadAndEyesLookAtExec  -  codel EXEC de HeadAndEyesLookAt
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneHeadAndEyesLookAtExec (CLONE_LOOK_AT_POINT *lookat, int *bilan)
{
  CLONE_LOOK_AT_POINT lookatTmp=*lookat;

  /* let's move */
  if (!cloneMove(&lookatTmp.pos, CLONE_MOVES_BOTH, lookatTmp.frame, bilan)) 
    return ETHER;

  return EXEC;
}


/*------------------------------------------------------------------------
 * LookTrack
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_clone_CANNOT_READ_POSTER
 */

#ifdef ISY 

static POSTER_ID trackPosterId;
static BOOL nothingToTrack;

/* cloneLookTrackStart  -  codel START de LookTrack
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneLookTrackStart (CLONE_TRACK *track, int *bilan)
{
  if (posterFind(track->posterName, &trackPosterId) == ERROR) {
    *bilan = S_clone_CANNOT_READ_POSTER;
    return ETHER;
  }
  nothingToTrack = FALSE;
  return EXEC;
}


/* cloneLookTrackExec  -  codel EXEC de LookTrack
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneLookTrackExec (CLONE_TRACK *track, int *bilan)
{
  ISY_RESULT_POSTER_STR  isyPosStr_cm;
  CLONE_POS pos;

  /* get target pos */
  if (cloneISY_RESULT_POSTER_STRPosterRead (trackPosterId, 
					    &isyPosStr_cm) == ERROR) {
    *bilan = S_clone_CANNOT_READ_POSTER;
    return ETHER;
  }
  if (isyPosStr_cm.result.number_of_faces <= 0) {
    if (!nothingToTrack)
      printf ("clone : nothing to track\n");
    nothingToTrack = TRUE;
    return EXEC;
  }

  nothingToTrack = FALSE;

  /* cm -> m */
  pos.x = (- isyPosStr_cm.result.position_x[0])/100.0;
  pos.y = (isyPosStr_cm.result.position_y[0])/100.0;
  pos.z = (isyPosStr_cm.result.position_z[0])/100.0;

  /* let's move */
  if (!cloneMove(&pos, track->motion, CLONE_ROBOT_FRAME, bilan))
    return ETHER;
  
  return EXEC;
}

#endif /* ISY */

/* ----------------------------------------------------------------------
 *
 * cloneMove - Moves the clone
 *
 *
 *
 */
static BOOL cloneMove(CLONE_POS *pos, CLONE_MOTION motion, CLONE_FRAME frame, 
		      int *bilan) 
{
  double headPercent;
  CLONE_POS posCorr;

  /* update screen pos */
  if (!cloneUpdateClonePositions (bilan))
    return FALSE;

  switch (frame) {
  case CLONE_CLONE_FRAME:
    posCorr=*pos;
    break;

  case CLONE_ROBOT_FRAME:
   posCorr.x = robot_to_clone[0] * pos->x + robot_to_clone[1] * pos->y + 
      robot_to_clone[2] * pos->z + robot_to_clone[3];
    posCorr.y = robot_to_clone[4] * pos->x + robot_to_clone[5] * pos->y + 
      robot_to_clone[6] * pos->z + robot_to_clone[7];
    posCorr.z = robot_to_clone[8] * pos->x + robot_to_clone[9] * pos->y + 
      robot_to_clone[10] * pos->z + robot_to_clone[11];
#if VERBOSE>0
    printf ("robot_to_clone (m): %5.2f %5.2f %5.2f -> %5.2f %5.2f %5.2f\n",
	    pos->x, pos->y, pos->z, 
	    posCorr.x, posCorr.y, posCorr.z);
#endif
    break;

  case CLONE_WORLD_FRAME:
    posCorr.x = world_to_clone[0] * pos->x + world_to_clone[1] * pos->y + 
      world_to_clone[2] * pos->z + world_to_clone[3];
    posCorr.y = world_to_clone[4] * pos->x + world_to_clone[5] * pos->y + 
      world_to_clone[6] * pos->z + world_to_clone[7];
    posCorr.z = world_to_clone[8] * pos->x + world_to_clone[9] * pos->y + 
      world_to_clone[10] * pos->z + world_to_clone[11];
#if VERBOSE>0
    printf ("world_to_clone (m): %5.2f %5.2f %5.2f -> %5.2f %5.2f %5.2f\n",
	    pos->x, pos->y, pos->z, 
	    posCorr.x, posCorr.y, posCorr.z);
#endif
    break;

  default:
    *bilan = S_clone_BAD_PARAMS;
    return FALSE;
  }

  if(motion == CLONE_MOVES_BOTH)
    headPercent = 0.6;
  else
    headPercent = 1;

  printf ("clone look at (m):  %5.2f  %5.2f  %5.2f  (%s frame)\n", 
	  posCorr.x, posCorr.y, posCorr.z,
	  frame == CLONE_ROBOT_FRAME ? "ROBOT" : 
	  frame == CLONE_CLONE_FRAME ? "CLONE" : "WORLD");
  
  
  /* add correction */
  cloneFrameCorrection(&posCorr, frame);

  printf ("clone look at (m):  %5.2f  %5.2f  %5.2f  (%s frame) corrected\n", 
	  posCorr.x, posCorr.y, posCorr.z,
	  frame == CLONE_ROBOT_FRAME ? "ROBOT" : 
	  frame == CLONE_CLONE_FRAME ? "CLONE" : "WORLD");

  /* Send the 3D point to look at to Minim in clone frame and in cm */
  if (motion == CLONE_MOVES_HEAD ||
      motion == CLONE_MOVES_BOTH)
    Minim_head_direction3D(1, headPercent, 
			   100.0*posCorr.x, 100.0*posCorr.y, 100.0*posCorr.z); 
  
  if (motion == CLONE_MOVES_EYES ||
      motion == CLONE_MOVES_BOTH)
    Minim_eyes_direction3D(0.6, 
			   100.0*posCorr.x, 100.0*posCorr.y, 100.0*posCorr.z, 1);

  return TRUE;
}


/* ----------------------------------------------------------------------
 *
 * transfPt
 *
 *
 */

static void transfPt(CLONE_POS *oldPos, POM_EULER_V *teul, CLONE_POS *newPos)
{
  T3D t3d;
  double x,y,z;
  double *matrix;

  pomEulerVToT3D(teul, &t3d);
  t3dConvertTo(T3D_MATRIX, &t3d);
  matrix = t3d.matrix.matrix;

  x = matrix[0] * oldPos->x +
      matrix[1] * oldPos->y +
      matrix[2] * oldPos->z +
      matrix[3];
  y = matrix[4] * oldPos->x +
      matrix[5] * oldPos->y +
      matrix[6] * oldPos->z +
      matrix[7];
  z = matrix[8] * oldPos->x +
      matrix[9] * oldPos->y +
      matrix[10] * oldPos->z +
      matrix[11];

  newPos->x = x;
  newPos->y = y;
  newPos->z = z;
}

/* ----------------------------------------------------------------------
 *
 * changePtFrame
 *
 *
 */
static BOOL changePtFrame(CLONE_POS *oldPos, CLONE_FRAME oldFrame,
			  CLONE_POS *newPos, CLONE_FRAME newFrame)
{
  switch(oldFrame) {
  case CLONE_CLONE_FRAME:
    switch(newFrame) {
    case CLONE_CLONE_FRAME:
      *newPos = *oldPos;
      break;
    case CLONE_ROBOT_FRAME:
      transfPt(oldPos, &t3d_c2m, newPos);
      break;
    case CLONE_WORLD_FRAME:
      transfPt(oldPos, &t3d_c2o, newPos);
      break;
    default:
      return FALSE;
    }
    break;
  case CLONE_ROBOT_FRAME:
    switch(newFrame) {
    case CLONE_CLONE_FRAME:
      transfPt(oldPos, &t3d_m2c, newPos);
      break;
    case CLONE_ROBOT_FRAME:
      *newPos = *oldPos;
      break;
    case CLONE_WORLD_FRAME:
      transfPt(oldPos, &t3d_m2o, newPos);
      break;
    default:
      return FALSE;
    }
    break;
  case CLONE_WORLD_FRAME:
    switch(newFrame) {
    case CLONE_CLONE_FRAME:
      transfPt(oldPos, &t3d_o2c, newPos);
      break;
    case CLONE_ROBOT_FRAME:
      printf ("clone : WORLD -> ROBOT frame not implemented\n");
/*       transfPt(oldPos, &t3d_o2m, newPos); */
      return FALSE;
      break;
    case CLONE_WORLD_FRAME:
      *newPos = *oldPos;
      break;
    default:
      return FALSE;
    }
    break;
  default:
    return FALSE;
  }
  return TRUE;
}

/* ----------------------------------------------------------------------
 *
 * cloneFrameCorrection   XXXXXX
 *
 *
 */
static void cloneFrameCorrection(CLONE_POS *pos, CLONE_FRAME frame) 
{
  static POM_EULER_V t3d_corr;

#if VERBOSE>0
  if (firsttime)
    printf ("clone corr : %5.2f %5.2f %5.2f  -> ",
	    pos->x, pos->y, pos->z);
#endif

  /* switch to clone frame */
  changePtFrame(pos, frame, pos, CLONE_CLONE_FRAME);

  /* correction */
  t3d_corr.euler.yaw = SDI_F->lookCorrection.rz;
  t3d_corr.euler.pitch = SDI_F->lookCorrection.ry;
  t3d_corr.euler.roll = SDI_F->lookCorrection.rx;
  t3d_corr.euler.x = SDI_F->lookCorrection.x;
  t3d_corr.euler.y = SDI_F->lookCorrection.y;
  t3d_corr.euler.z = SDI_F->lookCorrection.z;

  transfPt(pos, &t3d_corr, pos);

  /* convert back frame */
  changePtFrame(pos, CLONE_CLONE_FRAME, pos, frame);

#if VERBOSE>0
  if (firsttime)
    printf ("%5.2f %5.2f %5.2f\n",
	    pos->x, pos->y, pos->z);
#endif
}

/*------------------------------------------------------------------------
 * LookAtSpeechObjs
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* cloneLookAtSpeechObjsExec  -  codel EXEC de LookAtSpeechObjs
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneLookAtSpeechObjsExec (int *bilan)
{
  /* start */
  printf ("cloneLookAtSpeechObjsExec: not yet\n");

  
  /* update screen pos */
  if (!cloneUpdateClonePositions (bilan))
    return ETHER;

  return ETHER;
}


/*------------------------------------------------------------------------
 * AddSpeechObj
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_clone_BAD_PARAMS
 */

/* cloneAddSpeechObjExec  -  codel EXEC de AddSpeechObj
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneAddSpeechObjExec (CLONE_SPEECH_OBJ *speechObj, int *bilan)
{
  printf ("cloneAddSpeechObjExec: not yet\n");
  return ETHER;
}

/*------------------------------------------------------------------------
 * RemoveSpeechObj
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_clone_BAD_PARAMS
 */

/* cloneRemoveSpeechObjExec  -  codel EXEC de RemoveSpeechObj
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneRemoveSpeechObjExec (GEN_STRING256 *speechObjName, 
					 int *bilan)
{
  printf ("cloneRemoveSpeechObjExec: not yet\n");
  return ETHER;
}

/*------------------------------------------------------------------------
 * StereotypeMotion
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* cloneStereotypeMotionStart  -  codel START de StereotypeMotion
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneStereotypeMotionStart (CLONE_STEREOTYPE *stereotype, int *bilan)
{
  /* ... add your code here ... */
  return ETHER;
}

/* cloneStereotypeMotionExec  -  codel EXEC de StereotypeMotion
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneStereotypeMotionExec (CLONE_STEREOTYPE *stereotype, int *bilan)
{
  /* ... add your code here ... */
  return ETHER;
}

/* cloneStereotypeMotionEnd  -  codel END de StereotypeMotion
   Retourne:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT cloneStereotypeMotionEnd (CLONE_STEREOTYPE *stereotype, int *bilan)
{
  /* ... add your code here ... */
  return ETHER;
}



